#include <QPainter>
#include <QPaintEvent>
#include <QTextBlock>
#include "CodeWidget.hxx"

// --- CodeWidget: constructors and deconstructors ---

CodeWidget::CodeWidget(QWidget *parent)
: QPlainTextEdit(parent), _area(new LinesArea(this))
{
    setupActions();
    updateAreaWidth(0);
    highlightLine();
}

CodeWidget::CodeWidget(const QString& text, QWidget *parent)
: QPlainTextEdit(text, parent), _area(new LinesArea(this))
{
    setupActions();
    updateAreaWidth(0);
    highlightLine();
}

CodeWidget::~CodeWidget()
{
    delete _area;
}

// --- CodeWidget: public methods ---

void CodeWidget::areaPaintEvent(QPaintEvent *event)
{
    QPainter painter(_area);
    QTextBlock block = firstVisibleBlock();
    qint32 block_num = block.blockNumber();
    qint32 top = blockBoundingGeometry(block).translated(contentOffset()).top();
    qint32 bottom = top + blockBoundingRect(block).height();

    painter.fillRect(event->rect(), Qt::lightGray);

    while (block.isValid() && top <= event->rect().bottom())
    {
        if (block.isVisible() && bottom >= event->rect().top())
        {
            QString number = QString::number(block_num + 1);
            painter.setPen(Qt::black);
            painter.drawText(0, top, _area->width(), fontMetrics().height(), Qt::AlignRight, number);
        }

        block = block.next();
        top = bottom;
        bottom = top + blockBoundingRect(block).height();
        ++block_num;
    }
}

qint32 CodeWidget::areaWidth() const
{
    qint32 digits = 1;
    qint32 max = qMax(1, blockCount());
    qint32 space = 3;

    while (max >= 10)
    {
        max /= 10;
        ++digits;
    }

    space += fontMetrics().width(QLatin1Char('9')) * digits;

    return space;
}

// --- CodeWidget: protected methods ---

void CodeWidget::resizeEvent(QResizeEvent *event)
{
    QPlainTextEdit::resizeEvent(event);
    _area->setGeometry(QRect(contentsRect().left(), contentsRect().top(), areaWidth(), contentsRect().height()));
}

void CodeWidget::setupActions()
{
    connect(this, SIGNAL(blockCountChanged(qint32)),
            this, SLOT(updateAreaWidth(qint32)));
    connect(this, SIGNAL(updateRequest(QRect, qint32)),
            this, SLOT(updateArea(QRect, qint32)));
    connect(this, SIGNAL(cursorPositionChanged()),
            this, SLOT(highlightLine()));
}

// --- CodeWidget: protected slots ---

void CodeWidget::updateAreaWidth(qint32 /*block_count*/)
{
    setViewportMargins(areaWidth(), 0, 0, 0);
}

void CodeWidget::highlightLine()
{
    QList<QTextEdit::ExtraSelection> selections;

    if (!isReadOnly())
    {
        QTextEdit::ExtraSelection sel;
        QColor color = QColor(Qt::yellow).lighter(160);

        sel.format.setBackground(color);
        sel.format.setProperty(QTextFormat::FullWidthSelection, true);
        sel.cursor = textCursor();
        sel.cursor.clearSelection();
        selections.append(sel);
    }

    setExtraSelections(selections);
}

void CodeWidget::updateArea(const QRect& rect, qint32 dy)
{
    if (dy)
        _area->scroll(0, dy);
    else
        _area->update(0, rect.y(), _area->width(), rect.height());

    if (rect.contains(viewport()->rect()))
        updateAreaWidth(0);
}

// --- LinesArea: constructors and deconstructors ---

LinesArea::LinesArea(CodeWidget *parent)
: QWidget(parent), _parent(parent)
{
}

LinesArea::~LinesArea()
{
    _parent = 0;
}

// --- LinesArea: public methods ---

QSize LinesArea::sizeHint() const
{
    return QSize(_parent->areaWidth(), 0);
}

// --- LinesArea: protected methods ---

void LinesArea::paintEvent(QPaintEvent *event)
{
    _parent->areaPaintEvent(event);
}
